import numpy as np
import matplotlib
from matplotlib import colors, animation, pyplot as plt, rcParams
# from PIL import Image
from .utils import isinteger, within_0_1, within_0_255, isnotebook
from .config import config
from .flow import flow_to_color
import warnings
import logging
from datetime import datetime
import os
import copy
from numbers import Number


logger = logging.getLogger('TorchShow')

IMAGENET_MEAN = [0.485, 0.456, 0.406]
IMAGENET_STD = [0.229, 0.224, 0.225]


def create_color_map(N=256, normalized=False):
    
    def bitget(byteval, idx):
        return (byteval & (1 << idx)) != 0

    dtype = 'float32' if normalized else 'uint8'
    cmap = np.zeros((N, 3), dtype=dtype)
    for i in range(N):
        r = g = b = 0
        c = i
        for j in range(8):
            r = r | (bitget(c, 0) << 7-j)
            g = g | (bitget(c, 1) << 7-j)
            b = b | (bitget(c, 2) << 7-j)
            c = c >> 3

        cmap[i] = np.array([r, g, b])

    cmap = cmap/255 if normalized else cmap
    return cmap

def set_window_title(fig, title):
    """
    Set the title of the figure window (effective when using a interactive backend.)
    """
    try:
        fig.canvas.set_window_title(title)
    except:
        try:
            fig.canvas.manager.set_window_title(title)
        except:
            logger.info("Seting window title failed.")
    

def imshow(ax, vis, alpha=None, extent=None, show_rich_info=True):
    max_rows, max_cols = vis['raw'].shape[:2]
    
    def format_coord(x, y):
        """
        We display x-y coordinate as integer.
        """
        col = int(x + 0.5)
        row = int(y + 0.5)
        if not (0<=col<max_cols and 0<=row<max_rows):
            return ""
        return 'Mode='+vis['mode'] + ', Shape='+vis['shape'] + ', X=%1d, Y=%1d' % (col, row)
    
    def prepare_data(data):
        try:
            data[0]
        except (TypeError, IndexError):
            data = [data]
        return data
    
    def get_cursor_data(event):
        col = int(event.xdata + 0.5)
        row = int(event.ydata + 0.5)
        if not (0<=col<max_cols and 0<=row<max_rows):
            return ""
        raw_data = prepare_data(vis['raw'][row, col])
        disp_data = prepare_data(vis['disp'][row, col])
        
        return raw_data, disp_data
    
    def format_cursor_data(data):
        try:
            raw, disp = data
        except:
            return ""
        raw_data_str = ', '.join('{:0.3g}'.format(item) for item in raw
                                 if isinstance(item, Number))
        disp_data_str = ', '.join('{:0.3g}'.format(item) for item in disp
                                 if isinstance(item, Number))
        return "Raw=[{}], Disp=[{}]".format(raw_data_str, disp_data_str)
        # return "testest [" + str(data) + "]"
    
    artist = ax.imshow(vis['disp'], alpha=alpha, extent=extent, **vis['plot_cfg'])
    
    if show_rich_info:
        artist.get_cursor_data = get_cursor_data
        artist.format_cursor_data = format_cursor_data
        ax.format_coord = format_coord


def display_plt(vis_list, **kwargs):
    nrows = len(vis_list)
    ncols = max([len(l) for l in vis_list])
    show_axis = kwargs.get('show_axis', False)
    tight_layout = kwargs.get('tight_layout', True)
    suptitle = kwargs.get('suptitle', None)
    axes_title = kwargs.get('axes_title', None)
    # show_title = kwargs.get('show_title', False)
    # title_pattern = kwargs.get('title_pattern', "{img_id}")
    figsize = kwargs.get('figsize', None)
    dpi = kwargs.get('dpi', None)
    title_namespace = {}
    
    fig = plt.figure(figsize=figsize, dpi=dpi)
    axes = fig.subplots(nrows=nrows, ncols=ncols, squeeze=False)
    show_rich_info = config.get('show_rich_info')
    # fig, axes = plt.subplots(nrows=nrows, ncols=ncols, squeeze=False)
    set_window_title(fig, 'TorchShow')
    if suptitle:
        fig.suptitle(suptitle)
    
    for i, plots_per_row in enumerate(vis_list):
        for j, vis in enumerate(plots_per_row):
            # axes[i, j].imshow(vis, **plot_cfg)
            imshow(axes[i,j], vis, show_rich_info=show_rich_info)
            title_namespace["img_id"] = i*ncols+j
            title_namespace["img_id_from_1"] = title_namespace["img_id"] + 1
            title_namespace["row"] = i
            title_namespace["column"] = j
            if axes_title is not None:
                axes[i, j].set_title(axes_title.format(**title_namespace))
    
    # Delete empty axes            
    for ax in axes.ravel():
        if not ax.has_data():
            fig.delaxes(ax)
           
    if not show_axis:
        for ax in axes.ravel():
            ax.axis('off')
            
    if tight_layout:
        fig.tight_layout()
    
    if kwargs.get('save', False):
        file_path = kwargs.get('file_path', None)
        if file_path is None:
            os.makedirs('_torchshow', exist_ok=True)
            cur_time = datetime.now().strftime("%Y-%m-%d_%H-%M-%S-%f")
            file_path = '_torchshow/'+cur_time+'.png'
        dirname = os.path.dirname(file_path)
        if dirname != '':
            os.makedirs(dirname, exist_ok=True)
        fig.savefig(file_path, bbox_inches = 'tight', pad_inches=0)
        plt.close(fig)
    else: # If the image is saved by ts.save() it will not call plt.show()
        if not isnotebook():
            plt.show()
    

def overlay_plt(vis_list, alpha, save_as, extent, **kwargs):
    show_axis = kwargs.get('show_axis', False)
    tight_layout = kwargs.get('tight_layout', True)
    suptitle = kwargs.get('suptitle', None)
    axes_title = kwargs.get('axes_title', None)
    
    figsize = kwargs.get('figsize', None)
    dpi = kwargs.get('dpi', None)
    fig = plt.figure(figsize=figsize, dpi=dpi)
    ax = fig.add_subplot()
    set_window_title(fig, 'TorchShow')
    if suptitle:
        fig.suptitle(suptitle)
    
    assert len(vis_list) == len(alpha)
    
    for vis, a in zip(vis_list, alpha):
        imshow(ax, vis, alpha=a, extent=extent, show_rich_info=False)
        
    if not show_axis:
        ax.axis('off')
            
    if tight_layout:
        fig.tight_layout()
        
    if save_as != None:
        assert isinstance(save_as, str)
        dirname = os.path.dirname(save_as)
        if dirname!='':
            os.makedirs(dirname, exist_ok=True)
        fig.savefig(save_as, bbox_inches = 'tight', pad_inches=0)
        plt.close(fig)
    else:
        if not isnotebook():
            plt.show()
            
        
        

def animate_plt(video_list, **kwargs):
    """
    : video_list: [t, row, col, image]
    """
    nrows = len(video_list[0])
    ncols = max([len(l) for l in video_list[0]])
    
    show_axis = kwargs.get('show_axis', False)
    tight_layout = kwargs.get('tight_layout', True)
    suptitle = kwargs.get('suptitle', None)
    # show_title = kwargs.get('show_title', False)
    # title_pattern = kwargs.get('title_pattern', "{img_id}")
    # title_namespace = {}
    figsize = kwargs.get('figsize', None)
    dpi = kwargs.get('dpi', None)
    
    fig = plt.figure(figsize=figsize, dpi=dpi)
    axes = fig.subplots(nrows=nrows, ncols=ncols, squeeze=False)
    # fig, axes = plt.subplots(nrows=nrows, ncols=ncols, squeeze=False)
    set_window_title(fig, 'TorchShow')
    if suptitle:
        fig.suptitle(suptitle)

    plots = []
    
    # Initialization
    for i, plots_per_row in enumerate(video_list[0]):
        for j, vis in enumerate(plots_per_row):
            if vis is not None:
                plot = axes[i, j].imshow(vis['disp'], **vis['plot_cfg'])
                plots.append(plot)
            else:
                plots.append(None)
    
    # Delete empty axes
    for ax in axes.ravel():
        if not ax.has_data():
            fig.delaxes(ax)
    
    if not show_axis:
        for ax in axes.ravel():
            ax.axis('off')
            
    if tight_layout:
        fig.tight_layout()            
    
    def run(frames_at_t):
        for i, plots_per_row in enumerate(frames_at_t):
            for j, vis in enumerate(plots_per_row):
                # axes[i, j].imshow(vis, **plot_cfg)
                if vis is not None:
                    # axes[i,j].figure.canvas.draw()
                    plots[i*ncols+j].set_data(vis['disp'])
        fig.canvas.draw()
        return plots
           
    # if tight_layout:
    #     fig.tight_layout()
    
    ani = animation.FuncAnimation(fig, run, video_list, blit=True, interval=5, repeat=True)
    
    if not config.get('inline'):
        plt.show()
    return ani


def auto_unnormalize_image(x):
    all_int = isinteger(np.unique(x)).all()
    range_0_1 = within_0_1(x)
    range_0_255 = within_0_255(x)
    has_negative = (x.min() < 0)
    
    if has_negative:
        logger.debug('Detects input has negative values, auto rescaling input to 0-1.')
        return rescale_0_1(x)
    if range_0_255 and all_int and (not range_0_1): # if image is all integer and between 0 - 255. Normalize it to 0-1.
        logger.debug('Detects input are all integers within range 0-255. Divided all values by 255.')
        return x / 255.
    if range_0_1:
        logger.debug('Inputs already within 0-1, no unnormalization is performed.')
        return x
    logger.debug('Auto rescaling input to 0-1.')
    return rescale_0_1(x)


def rescale_0_1(x):
    """
    Rescaling tensor to 0-1 using min-max normalization
    """
    return (x - x.min()) / (x.max() - x.min())

        
def unnormalize_with_mean_and_std(x, mean, std):
    """
    General Channel-wise mean std unnormalization. Expect input to be (H, W, C)
    """
    x = x.copy()
    assert (len(x.shape) == 3), "Unnormalization only support (H, W, C) format input, got {}".format(x.shape)
    C = x.shape[-1]
    assert (len(mean) == C) and (len(std) == C), "Number of mean and std values must equals to number of channels."
    
    for i in range(C):
        x[:,:,i] = x[:,:,i] * std[i] + mean[i]
        
    return x
        
        
def vis_image(x, unnormalize='auto', **kwargs):
    """
    : x: ndarray (H, W, 3).
    : unnormalize: 'auto', 'imagenet', 'imagenet_scaled'
    : mean: image mean for unnormalization.
    : std: image std for unnormalization
    : display: whether to display image using matplotlib
    """
    vis = dict()
    vis['raw'] = copy.deepcopy(x)
    vis['shape'] = str(x.shape)
    shape = x.shape
    ndim = len(shape)
    assert ndim == 3, "vis_image only support 3D array in (H, W, C) format."
    
    user_mean = config.get('image_mean')
    user_std = config.get('image_std')
    
    if (user_mean is not None) or (user_std is not None):
        if user_mean == None:
            user_mean = [0] * x.shape[-1] # Initialize mean to 0 if not specified.
        if user_std == None:
            user_std = [1.] * x.shape[-1] # Initialize std to 1 if not specified.
        x = unnormalize_with_mean_and_std(x, user_mean, user_std)
        
    elif unnormalize=='auto':
        x = auto_unnormalize_image(x)
        
    elif unnormalize=='imagenet':
        if (x.max() > 2.66) or (x.min() < -2.17): 
            # A quick validation to check if the image was normalized to 0-1 
            # before substracting imagenet mean and std
            x = x / 255.
        x = unnormalize_with_mean_and_std(x, IMAGENET_MEAN, IMAGENET_STD)
        
    else:
        raise NotImplementedError("Unsupported unnormalization profile \"{}\"".format(unnormalize))
    
    assert x is not None
    
    if config.get('color_mode') == 'bgr':
        x = x[:,:,::-1]
        vis['mode'] = 'Image(BGR)'
    else:
        vis['mode'] = 'Image(RGB)'
    
    plot_cfg = dict()
    
    vis['disp'] = x
    vis['plot_cfg'] = plot_cfg
    return vis

    
def vis_flow(x, **kwargs):
    vis = dict()
    vis['raw'] = copy.deepcopy(x)
    vis['shape'] = str(x.shape)
    x = flow_to_color(x)
    plot_cfg = dict()
    vis['disp'] = x
    vis['plot_cfg'] = plot_cfg
    vis['mode'] = 'Flow'
    return vis


def vis_grayscale(x, **kwargs):
    vis = dict()
    assert (len(x.shape) == 3) and (x.shape[-1] == 1)
    x = np.squeeze(x, -1)
    vis['raw'] = copy.deepcopy(x)
    vis['shape'] = str(x.shape)
    # rescale to [0-1]
    if not within_0_1(x):
        warnings.warn('Original input range is not 0-1 when using grayscale mode. Auto-rescaling it to 0-1 by default.')
        x = rescale_0_1(x)
    vis['disp'] = x
    cmap = kwargs.get("cmap", "gray")
    plot_cfg = dict(cmap=cmap)
    
    if isinteger(np.unique(x)).all():
        plot_cfg['interpolation'] = 'nearest'
        vis['mode'] = 'Binary'
    else:
        vis['mode'] = 'Gray'
    vis['plot_cfg'] = plot_cfg
    
    return vis


def vis_categorical_mask(x, max_N=256, **kwargs):
    assert (len(x.shape) == 3) and (x.shape[-1] == 1)
    assert isinteger(np.unique(x)).all(), "Input has to contain only integers in categorical mask mode."
    vis = dict()
    
    x = np.squeeze(x, -1)
    vis['raw'] = copy.deepcopy(x)
    vis['shape'] = str(x.shape)
    N = int(x.max()) + 1
    
    if x.max() > max_N:
        warnings.warn('The maximum value in input is {} which is greater than the default max_N ({}). Automatically adjust max_N to {}.'.format(x.max(), max_N, x.max()))
        max_N = x.max() + 1
        
    color_list = create_color_map(N=max_N, normalized=True)
    color_list = np.concatenate([color_list, np.ones((max_N, 1)).astype(np.float32)], axis=1)
    
    if x.min() < 0:
        warnings.warn('Input has negative value when trying to visualize as categorical mask, which will all be converted to -1 and displayed in white.')
        x[x<0] = -1 # Map all negative value to -1
        color_list = np.concatenate([np.ones((1,4)).astype(np.float32), color_list], axis=0) # appending an extra value.
        N = N + 1
    
    cmap = colors.ListedColormap(color_list, N=N)
    
    x = cmap(x.astype(int), alpha=None, bytes=True)[:,:,:3]
    # print(x.shape)
    plot_cfg = dict( interpolation="nearest")
    
    vis['disp'] = x
    vis['plot_cfg'] = plot_cfg
    vis['mode'] = 'Categorical'
    return vis


if __name__ == "__main__":
    print(create_color_map())



import sys
import torchshow as ts
import torch 
import torchvision
import numpy as np
import logging
from PIL import Image

def read_flow(filename):
    with open(filename, 'rb') as f:
        magic = np.fromfile(f, np.float32, count=1)
        if 202021.25 != magic:
            print('Magic number incorrect. Invalid .flo file')
        else:
            w = np.fromfile(f, np.int32, count=1)[0]
            h = np.fromfile(f, np.int32, count=1)[0]
            print('Reading %d x %d flo file' % (w, h))
            data = np.fromfile(f, np.float32, count=2*w*h)
            # Reshape data into 3D array (columns, rows, bands)
            data2D = np.resize(data, (h, w, 2))
        return data2D
            

def test(section):
    rgb_img = torch.rand((3, 100, 100))
    if section <= 1:
        print("1.1 Single 3-channel image between 0-1")
        rgb_img = torch.rand((3, 100, 100))
        print(rgb_img.min(), rgb_img.max())
        ts.show(rgb_img)
        ts.save(rgb_img)

        print("1.2 Single 3-channel image between 0-255")
        rgb_img_2 = rgb_img * 100
        print(rgb_img_2.min(), rgb_img_2.max())
        ts.show(rgb_img_2)
        ts.save(rgb_img_2)

        print("1.3 Single 3-channel image with value larger than 255")
        rgb_img_3 = rgb_img * 300
        print(rgb_img_3.min(), rgb_img_3.max())
        ts.show(rgb_img_3)
        ts.save(rgb_img_3)

        print("1.4 Single 3-channel image with value smaller than 0")
        rgb_img_4 = rgb_img - 0.5
        print(rgb_img_4.min(), rgb_img_4.max())
        ts.show(rgb_img_4)
        ts.save(rgb_img_4)

        print("1.5 Single 3-channel image with value smaller than 0 and larger than 255")
        rgb_img_5 = rgb_img * 500 - 100
        print(rgb_img_5.min(), rgb_img_5.max())
        ts.show(rgb_img_5)
        ts.save(rgb_img_5)

    gray_img = torch.rand((1, 100, 100))
    category_mask = np.array(Image.open('test_data/example_category_mask.png'))
    depth_image = np.array(Image.open('test_data/example_depth.png')) / 1000.0
    if section <=2:
        print("2.1 Single 1-Channel image between 0-1")
        print(gray_img.min(), gray_img.max())
        ts.show(gray_img)
        ts.save(gray_img)

        print("2.2 Single 1-Channel image between 0-255")
        gray_img_2 = gray_img * 100
        print(gray_img_2.min(), gray_img_2.max())
        ts.show(gray_img_2)
        ts.save(gray_img_2)

        print("2.3 Single 1-Channel image with value smaller than 0 and larger than 255")
        gray_img_3 = gray_img *  500 - 100
        print(gray_img_3.min(), gray_img_3.max())
        ts.show(gray_img_3)
        ts.save(gray_img_3)

        print("2.4 Single 1-Channel image with binary value")
        gray_img_4 = torch.eye(100).unsqueeze(0)
        gray_img_4[:, 20:40, 20:40] = 1
        gray_img_4[:, 60:80, 60:80] = 1 
        print(gray_img_4.unique())
        ts.show(gray_img_4)
        ts.save(gray_img_4)

        print("2.5 Single 1-Channel image with integer value >= 0")
        gray_img_5 = torch.randint(0, 100, (1, 100, 100))
        print(gray_img_5.unique())
        ts.show(gray_img_5)
        ts.save(gray_img_5)

        print("2.6 Single 1-Channel image with both positive and negative integer value")
        gray_img_6 = torch.randint(-50, 100, (1, 100, 100))
        print(gray_img_6.unique())
        ts.show(gray_img_6)
        ts.save(gray_img_6)
        
        print("2.7 Single 1-Channel normal categorical mask")
        print(np.unique(category_mask))
        ts.show(category_mask)
        ts.save(category_mask)
        
        print("2.8 Single channel image with custom cmap")
        ts.show(depth_image, cmap="magma")
        ts.show(depth_image, cmap="inferno")
        ts.show(depth_image, cmap="viridis")
        ts.save(depth_image, cmap="magma")
        ts.save(depth_image, cmap="inferno")
        ts.save(depth_image, cmap="viridis")

    flow = read_flow("./test_data/example_flow.flo")
    if section <= 3:
        print("3.1 Single 2-Channel normal optical flow")
        print(flow.shape, flow.min(), flow.max())
        ts.show(flow)
        ts.save(flow)

        print("3.2 Single 2-Channel with random values")
        flow_2 = torch.rand((2, 100, 100)) * 200 - 100
        print(flow_2.min(), flow_2.max())
        ts.show(flow_2)
        ts.save(flow_2)

    img_n = torch.rand(16, 100, 100)
    if section <=4:
        print("4.1 Single n-channel tensor (n>3")
        print(img_n.min(), img_n.max())
        ts.show(img_n)
        ts.save(img_n)
        ts.show(img_n, ncols=3)
        ts.save(img_n, ncols=3)
        ts.show(img_n, nrows=5)
        ts.save(img_n, nrows=5)

        print("4.2 4D tensors")
        batch = torch.rand(8, 3, 100, 100) * 500 - 250
        print(batch.min(), batch.max())
        ts.show(batch)
        ts.save(batch)
        ts.show(batch, ncols=3)
        ts.save(batch, ncols=3)
        ts.show(batch, nrows=4)
        ts.save(batch, nrows=4)
        ts.show(batch, axes_title="Image ID: {img_id_from_1}")
        ts.show(batch, nrows=4, axes_title="{img_id}-{img_id_from_1}-{row}-{column}")
        ts.show(batch, nrows=4, axes_title="{img_id}-{img_id_from_1}-{row}-{column}", suptitle="Figure 1")
        
        print("4.3 List of tensors")
        list_tensors = list(batch)
        ts.show(list_tensors)
        ts.save(list_tensors)
        ts.show(list_tensors, ncols=3)
        ts.save(list_tensors, ncols=3)
        ts.show(list_tensors, nrows=4)
        ts.save(list_tensors, nrows=4)
        ts.show(list_tensors, nrows=3, ncols=5)
        ts.save(list_tensors, nrows=3, ncols=5)
        ts.show(list_tensors, axes_title="Image ID: {img_id_from_1}")
        ts.show(list_tensors, nrows=4, axes_title="{img_id}-{img_id_from_1}-{row}-{column}")
        ts.show(list_tensors, nrows=4, axes_title="{img_id}-{img_id_from_1}-{row}-{column}", suptitle="List of Tensors")

    if section <=5:
        print("5.1 Custom Layout")
        grid = [[rgb_img, gray_img, flow]]
        ts.show(grid)
        ts.save(grid)
        grid2 = [[rgb_img, gray_img], 
                 [flow]]
        ts.show(grid2)
        ts.save(grid2)

    if section <=6:
        print("6.1 Video Clip")
        video = torch.rand(16, 3, 100, 100)
        print(video.min(), video.max())
        ts.show_video(video)
        video2 = torch.rand(8, 1, 100, 100)
        video3 = torch.rand(13, 2, 100, 100)
        ts.show_video([[video, video2], 
                       [video3]])
        video4 = torch.rand(13,5, 100, 100)
        list_of_videos = [video, video2, video3, video, video2, video3]
        ts.show_video(list_of_videos)
        ts.show_video(list_of_videos, ncols=3)
        ts.show_video(list_of_videos, nrows=3)
        ts.show_video(list_of_videos, nrows=2, ncols=8)
        print("6.2 ts.show_video with image") 
        # This test produces unwanted results. Ignore it at this moment unless requested.
        ts.show_video(rgb_img)
        ts.show_video(flow)
        ts.show_video([[video, video2], 
                       [rgb_img, flow]], suptitle="Video Example")
        vis_list = ts.show_video([[video, video2], 
                       [rgb_img, flow]], display=False)
        print(len(vis_list), len(vis_list[0]), len(vis_list[0][0]))
        print("6.3 Video Clip Edge Cases")
        try:
            ts.show_video(video4)
        except Exception as e: # This should raise an error
            print(e)
        try:
            ts.show_video([[video, video2], 
                           [video3, video4]])
        except Exception as e: # This should raise an error
            print(e)

    if section <=7:
        print("7 Return vis_list if display=False")
        vis_list = ts.show(img_n, display=False)
        print(len(vis_list), len(vis_list[0]))
        vis_list = ts.show(img_n, display=False, nrows=3)
        print(len(vis_list), len(vis_list[0]))
 
    if section <=8:
        print("8 Change Unnormalization Presets")
        rgb_img_numpy = rgb_img.permute(1,2,0).numpy()
        def test_normalize(MEAN, STD):
            transform = torchvision.transforms.Normalize(MEAN, STD)
            rgb_img_0 = transform(rgb_img)
            ts.set_image_mean(MEAN)
            ts.set_image_std(STD)
            rgb_img_1 = ts.show(rgb_img_0, display=False)[0][0]['disp']
            assert np.allclose(rgb_img_1, rgb_img_numpy, atol=1e-7)  # will be False if atol=1e-8

        test_normalize([0.5, 0.5, 0.5], [0.5, 0.5, 0.5])
        test_normalize([0.485, 0.456, 0.406], [0.229, 0.224, 0.225])
        test_normalize([0.4851231531212364564, 0.4561231523135436, 0.406123412312452343], [0.2293453455673435, 0.22434531445342, 0.22534534472423])

    ts.set_image_mean([0,0,0])
    ts.set_image_std([1,1,1])

    if section <=9:
        print("9 Testing PIL Image")
        pil_rgb = Image.open("test_data/example_category_mask.png")
        pil_depth = Image.open("test_data/example_depth.png")
        ts.show([pil_rgb, pil_depth])
        ts.save([pil_rgb, pil_depth])
        
    if section <=10:
        print("10 Testing filename as input")
        ts.show(["test_data/example_image.jpeg", "test_data/example_depth.png", "test_data/example_category_mask.png"])
        try:
            ts.show("test_data/a_file_that_does_not_exists.jpg")
        except Exception as e :
            print(e)
            pass
        ts.show("test_data/example_image_rotated_by_exif.jpeg")
        ts.show("test_data/example_flow.flo")
        
    if section <= 11:
        print("11 Testing Overlay FUnction")
        ts.overlay(["test_data/example_rgb.jpg", "test_data/example_category_mask.png"], alpha=[0.8])
        ts.overlay(["test_data/example_rgb.jpg", "test_data/example_category_mask.png"], alpha=[0.5, 0.5])
        ts.overlay(["test_data/example_rgb.jpg", "test_data/example_category_mask.png", "test_data/example_flow.flo"], alpha=[0.5, 0.5])
        ts.overlay(["test_data/example_rgb.jpg", "test_data/example_category_mask.png"], save_as="_torchshow/test_overlay_save.png")
        # ts.overlay(["test_data/example_rgb.jpg", "test_data/example_category_mask.png"], alpha=[0.5, 1.5])

if __name__ == "__main__":
    if len(sys.argv) > 1:
        section = sys.argv[1]
    else:
        section = 0
    test(int(section))


import numpy as np

import logging
import warnings

from .visualization import vis_image, vis_grayscale, vis_categorical_mask, vis_flow, display_plt, animate_plt, overlay_plt
from .utils import isinteger, calculate_grid_layout, tensor_to_array

logging.basicConfig(level=logging.INFO)

logger = logging.getLogger('TorchShow')

vis_func_dict = dict(image=vis_image,
                     grayscale=vis_grayscale,
                     categorical_mask=vis_categorical_mask,
                     flow=vis_flow)

def save(x, path=None, **kwargs):
    show(x, save=True, file_path=path, **kwargs)


def show(x, display=True, **kwargs):
    vis_list = None
    
    x = tensor_to_array(x)

    if isinstance(x, (np.ndarray)):
        x = x.copy()
        nrows = kwargs.get('nrows', None)
        ncols = kwargs.get('ncols', None)
        channel_mode = kwargs.get('channel_mode', 'auto')
        if channel_mode == 'auto':
            if x.shape[-1] in [1,2,3]:
                channel_mode = 'channel_last'
            else:
                channel_mode = 'channel_first'
                
        if x.ndim == 4: # (N, C, H, W) like array
            if channel_mode == 'channel_first':
                N, _, H, W = x.shape
            elif channel_mode == 'channel_last':
                N, H, W, _ = x.shape
            
            nrows, ncols = calculate_grid_layout(N, H, W, nrows, ncols)
            assert (nrows * ncols >= N)
            vis_list = [list(x[i:i + ncols]) for i in range(0, N, ncols)] # vis_list is now an list of list

            
        elif x.ndim == 3: # (C, H, W) like array
            if channel_mode == 'channel_first': # C, H, W
                C, H, W = x.shape
            elif channel_mode == 'channel_last':
                H, W, C = x.shape
                
            if C <=3:
                vis_list = [[x]] # if C is in [1,2,3], visualize it as single image
            else: # when C is greater than 3 (e.g. feature maps), visualize it in grid layout
                if channel_mode == 'channel_last':
                    x = np.transpose(x, (2,0,1)) # Transpose to C, H, W because we will visualize each individual channel
                nrows, ncols = calculate_grid_layout(C, H, W, nrows, ncols)
                assert (nrows * ncols >= C)
                vis_list = [list(x[i:i + ncols]) for i in range(0, C, ncols)]
                
        elif x.ndim == 2: # (H, W)
            vis_list = [[x]]
            
        else:
            raise TypeError("Unsupported shape of numpy array {} .".format(x.shape))
        
    elif isinstance(x, list):
        if isinstance(x[0], np.ndarray): # if the input is list of images [img1, img2, ....]
            nrows = kwargs.get('nrows', None)
            ncols = kwargs.get('ncols', None)
            if (nrows is not None) or (ncols is not None): # If user specified the grid layout
                N = len(x)
                # Here we assume either `nrows` or `ncols` must given so we do not have to specify H, W
                nrows, ncols = calculate_grid_layout(N, None, None, nrows, ncols)
                assert (nrows * ncols >= N)
                vis_list = [list(x[i:i + ncols]) for i in range(0, N, ncols)] # vis_list is now a list of list
            else:    
                vis_list = [x] # Make it [[img1, img2, ...]] if `nrows` or `ncols` are not specified
        else:
            vis_list = x

    else:
        raise NotImplementedError("Does not support input type \"{}\"".format(type(x)))


    # vis_list:  list of list. Outer list is for rows and inner list is the images in each row.
    # e.g.[[img1, img2], 
    #      [img3, img4]]
    
    assert isinstance(vis_list, list)
    assert np.array([isinstance(l, list) for l in vis_list]).all() # Now the input should be list of list

    plot_list = []

    for row in vis_list: 
        list_per_row = []
        for img in row:
            vis = visualize(img, **kwargs)
            list_per_row.append(vis)
        plot_list.append(list_per_row)
    
    if display:
        display_plt(plot_list, **kwargs)
    else:
        return plot_list


def show_video(x, display=True, **kwargs):
    video_list = None
    
    x = tensor_to_array(x)

    if isinstance(x, (np.ndarray)):
        x = x.copy()
        assert x.ndim in [3,4], "only support 3D array (N, H, W) or 4D array (N, C, H, W) in video mode"
        print(x.shape)
        video_list = [[x]] # for a single video, make it [[vid]]
        
    elif isinstance(x, list):
        if isinstance(x[0], np.ndarray): # if the input is list of array [vid1, vid2, ...]
            nrows = kwargs.get('nrows', None)
            ncols = kwargs.get('ncols', None)
            if (nrows is not None) or (ncols is not None): # If user specified the grid layout
                N = len(x)
                # Here we assume either `nrows` or `ncols` must given so we do not have to specify H, W.
                nrows, ncols = calculate_grid_layout(N, None, None, nrows, ncols)
                assert (nrows * ncols >= N)
                video_list = [list(x[i:i + ncols]) for i in range(0, N, ncols)] # video_list is now a list of list
            else:    
                video_list = [x] # Make it [[vid1, vid2, ...]] if `nrows` or `ncols` are not specified
        else:
            video_list = x
            
    else:
        raise NotImplementedError("Does not support input type \"{}\"".format(type(x)))


    # video_list:  list of list. Outer list is for rows and inner list is the images in each row.
    # e.g.[[img1, img2], 
    #      [img3, img4]]
    
    assert isinstance(video_list, list)
    assert np.array([isinstance(l, list) for l in video_list]).all() # Now the input should be list of list

    video_length = max([len(vid) for l in video_list for vid in l])

    video_vis_list = [] # Reorganize frames into [t, row, col, img]

    for t in range(video_length): 
        frames_at_t = [] # [[frame_t_video1, frame_t_video2],
                         #  [frame_t_video3, frame_t_video4]]
        for row in video_list: 
            frames_at_t_per_row = [] # [frame_t_video1, frame_t_video2]
            for video in row:
                if t < len(video):
                    img = video[t]
                    vis = visualize(img, **kwargs)
                else:
                    vis = None

                frames_at_t_per_row.append(vis) # 
            frames_at_t.append(frames_at_t_per_row) # 
        video_vis_list.append(frames_at_t)
        
    if display:
        return animate_plt(video_vis_list, **kwargs)
    else:
        return video_vis_list

def overlay(x, alpha=None, extent=None, save_as=None, **kwargs):
    """ Overlay a list of inputs. Useful for comparing two images or putting masks over images.

    Args:
        x (list): a list of tensor-like.
        alpha (list) (Optional) : list of alpha for the overlay factor. Each alpha value could be a float or array-like.
        save_as (str) (Optional) : A filepath. If specified, save the result to this file.
    """
    assert isinstance(x, list) and len(x)>1, "ts.overlay() expect a list with at least two tensor-like as inputs."
    if alpha is None:
        alpha = [1] + [0.7] * (len(x)-1) # Default blending factor.
    if not isinstance(alpha, list):
        alpha = [alpha]
    
    if len(alpha) < len(x):
        alpha += [0.7] * (len(x) - len(alpha)) # fill the list of alpha so it has the same length as the list of tensor.
    
    vis_list = show(x, display=False, **kwargs)[0]
        
    if extent is None:
        h, w = vis_list[0]["disp"].shape[:2]
        extent = 0, w, 0, h
        
    return overlay_plt(vis_list, alpha=alpha, save_as=save_as, extent=extent, **kwargs)
            

def visualize(x, 
         mode='auto',
         auto_permute=True,
         **kwargs):

    assert isinstance(x, np.ndarray)
    
    shape = x.shape
    ndim = len(shape)
    assert ndim <= 3
    
    if auto_permute:
        if (ndim == 3) and (shape[0] in [1, 2, 3]): # For C, H, W kind of array.
            logger.debug('Detected input shape {} is in CHW format, TorchShow will automatically convert it to HWC format'.format(shape))
            x = np.transpose(x, (1,2,0))

    if ndim == 2:
        x = np.expand_dims(x, axis=-1)
    
    if mode=='auto':
        mode = infer_mode(x)
    
    vis_func = vis_func_dict.get(mode, None)
    
    if vis_func == None:
        raise ValueError("mode {} is not supported.".format(mode))
        
    return vis_func(x, **kwargs)
    

def infer_mode(x):
    shape = x.shape
    ndim = len(shape)
    if shape[-1] == 3:
        mode = 'image'
    elif shape[-1] == 2:
        mode = 'flow'
    elif shape[-1] == 1:
        if (x.min() >= 0) and (x.max() <= 1):
            mode = 'grayscale'
        elif isinteger(np.unique(x)).all(): # If values are all integer
            mode = 'categorical_mask'
        else:
            mode = 'grayscale'
    else:
        raise NotImplementedError("Does support auto infer for shape {} .".format(shape))
    logger.debug("Auto Infer: {}".format(mode))
    return mode




import numpy as np
import os
import sys
import importlib
import warnings
import numbers
import logging

_EXIF_ORIENT = 274

logger = logging.getLogger('TorchShow')

def isnumber(x):
    return isinstance(x, numbers.Number)


def isinteger(x):
    """
    Function to check if np array has integer. Work regardless of type.
    e.g. isinteger(np.array([0., 1.5. 1.])) >>> array([ True, False,  True], dtype=bool)
    
    """
    return np.equal(np.mod(x, 1), 0)
        

def within_0_1(x):
    return (x.min() >= 0) and (x.max() <= 1)


def within_0_255(x):
    return (x.min() >= 0) and (x.max() <= 255)


def calculate_grid_layout(N, img_H, img_W, nrow=None, ncol=None):
    """
    Function to calculate grid_layout
    """
    if (nrow != None): # `nrow` has higher priority than `ncol`.
        ncol = int(np.ceil(N / nrow))
    elif (ncol != None):
        nrow = int(np.ceil(N / ncol))
    else: # If both nrow and ncols are not set, perform automatic calculation.
        N_sqrt = np.sqrt(N)
        if img_H >= img_W:
            nrow = int(np.floor(N_sqrt))
            ncol = int(np.ceil(N/nrow))
        else:
            ncol = int(np.floor(N_sqrt))
            nrow = int(np.ceil(N/ncol))
    return nrow, ncol
    
    
def tensor_to_array(x):
    # Recursively perform tensor to array conversion
    # ====== PyTorch Tensor =======
    try:
        torch = import_if_not_yet("torch")
    except ImportError:
        pass
    else:
        if isinstance(x, torch.Tensor):
            x = x.detach().clone().cpu()
            if x.dtype in [torch.bfloat16, torch.float16, torch.bool]:
                logger.warning("The tensor with type \"{}\" was automatically converted to float32 for visualization.".format(x.dtype))
                x = x.float()
            return x.numpy()
    
    # ====== PIL Image =======
    try: 
        Image = import_if_not_yet(module_name="Image", package_name="PIL")
    except ImportError:
        pass
    else:
        if isinstance(x, Image.Image):
            return np.asarray(x)
    
    # ====== Numpy Array =======
    if isinstance(x, np.ndarray):
        return x.copy()
    
    # ====== Filename =======
    elif isinstance(x, str): # Handling filename
        if os.path.exists(x):
            if x.split('.')[-1].lower() == 'flo': # Handling optical flow files.
                return read_flow(x)
            else:
                image = read_image_PIL(x)
                return np.asarray(image)
        else:
            raise FileNotFoundError(f"{x} is not a file.")
    # ====== Recursively processing list of inputs
    elif isinstance(x, list):
        return [tensor_to_array(e) for e in x]
    else:
        raise TypeError('Found unsupported type ', type(x))


def isnotebook():
    # Check if running in ipython/jupyter notebook environment
    try:
        shell = get_ipython().__class__.__name__
        if shell == 'ZMQInteractiveShell':
            return True   # Jupyter notebook or qtconsole
        elif shell == 'TerminalInteractiveShell':
            return False  # Terminal running IPython
        else:
            return False  # Other type (?)
    except NameError:
        return False      # Probably standard Python interpreter
    
def import_if_not_yet(module_name, package_name=""):
    module_key = f"{package_name}.{module_name}".lstrip('.')
    if module_key not in sys.modules:
        return importlib.import_module(module_name, package_name)
    else:
        return sys.modules[module_key]

def read_image_PIL(filename):
    try: 
        Image = import_if_not_yet(module_name="Image", package_name="PIL")
    except ImportError:
        raise ImportError("TorchShow opens image files using PIL which is not yet installed.")
    else:
        image = Image.open(filename)
        mode = image.mode
        if mode == 'RGBA': # Convert RGBA to RGB since torchshow will handle 4 channel images differently.
            mode = 'RGB'
            
        """
        The following code is for correctly applying the exif orientation. 
        This code is modified based on https://github.com/facebookresearch/detectron2/blob/2b98c273b240b54d2d0ee6853dc331c4f2ca87b9/detectron2/data/detection_utils.py#L119
        """
        
        if not hasattr(image, "getexif"):
            return image

        try:
            exif = image.getexif()
        except Exception:  # https://github.com/facebookresearch/detectron2/issues/1885
            exif = None

        if exif is None:
            return image

        orientation = exif.get(_EXIF_ORIENT)
        method = {
            2: Image.FLIP_LEFT_RIGHT,
            3: Image.ROTATE_180,
            4: Image.FLIP_TOP_BOTTOM,
            5: Image.TRANSPOSE,
            6: Image.ROTATE_270,
            7: Image.TRANSVERSE,
            8: Image.ROTATE_90,
        }.get(orientation)

        if method is not None:
            warnings.warn(f"TorchShow has detected orientation information in the EXIF of file {filename} and the corresponding transformed has been applied. Please be aware of this issue if you want to load this file with PIL in your code.")
            image = image.transpose(method)
            
        return image.convert(mode)


def read_flow(filename):
    with open(filename, 'rb') as f:
        magic = np.fromfile(f, np.float32, count=1)
        if 202021.25 != magic:
            print('Magic number incorrect. Invalid .flo file')
        else:
            w = np.fromfile(f, np.int32, count=1)[0]
            h = np.fromfile(f, np.int32, count=1)[0]
            print('Reading %d x %d flo file' % (w, h))
            data = np.fromfile(f, np.float32, count=2*w*h)
            # Reshape data into 3D array (columns, rows, bands)
            data2D = np.resize(data, (h, w, 2))
        return data2D


if __name__ == '__main__':
    for N in range(1,100):
        row, col = calculate_grid_layout(N, 10, 10)
        print(N, row, col, row*col)


class Config():
    def __init__(self):
        self.config_dict = {'backend': 'matplotlib', # May support more backend in the future
                            'image_mean': None,
                            'image_std': None,
                            'color_mode': 'rgb',
                            'show_rich_info': True,
                            } 

    def set(self, key, value):
        if key in self.config_dict.keys():
            self.config_dict[key] = value
        
    def get(self, key):
        return self.config_dict.get(key, None)

config = Config()
        
def set_image_mean(mean: list):
    assert len(mean)==3
    config.set('image_mean', mean)
    
def set_image_std(std: list):
    assert len(std)==3
    config.set('image_std', std)
    
def set_color_mode(value='rgb'):
    assert value in ['rgb', 'bgr']
    config.set('color_mode', value)
    
def show_rich_info(flag: bool):
    if flag:
        config.set('show_rich_info', True)
    else:
        config.set('show_rich_info', False)